@@ -84,6 +84,7 @@
 //!   - Full of `unsafe`
 //!
 //! # Changelog
+//! - v0.2.3 - Implement `From<NewThreadPoolError>` and `From<TryScheduleError>` for `std::io::Error`.
 //! - v0.2.2 - Add `ThreadPool::join` and `ThreadPool::try_join`.
 //! - v0.2.1 - Improve test coverage.
 //! - v0.2.0
@@ -97,6 +98,7 @@
 //! - v0.1.0 - First release
 //!
 //! # TO DO
+//! - Make `join` and `try_join` work with `Arc<ThreadPool>`.
 //! - Log a warning when all threads panicked.
 //! - Update test coverage.
 //! - Add a public `respawn_threads` function.
@@ -109,65 +111,26 @@
 //! 1. Run `./release.sh`
 #![forbid(unsafe_code)]
 
+mod atomic_counter;
+
+use atomic_counter::AtomicCounter;
 use core::fmt::{Debug, Display, Formatter};
-use core::sync::atomic::{AtomicUsize, Ordering};
 use core::time::Duration;
 use std::error::Error;
+use std::io::ErrorKind;
 use std::sync::mpsc::{Receiver, RecvTimeoutError, SyncSender, TrySendError};
 use std::sync::{Arc, Mutex};
 use std::time::Instant;
 
-#[cfg(test)]
-pub static TEST_MAX_THREADS: AtomicUsize = AtomicUsize::new(100);
+#[cfg(feature = "testing")]
+#[doc(hidden)]
+pub static INTERNAL_MAX_THREADS: core::sync::atomic::AtomicUsize =
+    core::sync::atomic::AtomicUsize::new(usize::MAX);
 
 fn sleep_ms(ms: u64) {
     std::thread::sleep(Duration::from_millis(ms));
 }
 
-struct AtomicCounter {
-    next_value: AtomicUsize,
-}
-
-impl AtomicCounter {
-    pub fn new() -> Self {
-        Self {
-            next_value: AtomicUsize::new(0),
-        }
-    }
-    pub fn next(&self) -> usize {
-        self.next_value.fetch_add(1, Ordering::AcqRel)
-    }
-}
-
-#[test]
-fn atomic_counter() {
-    let counter = Arc::new(AtomicCounter::new());
-    assert_eq!(0, counter.next());
-    assert_eq!(1, counter.next());
-    assert_eq!(2, counter.next());
-}
-
-#[test]
-fn atomic_counter_many_readers() {
-    let receiver = {
-        let counter = Arc::new(AtomicCounter::new());
-        let (sender, receiver) = std::sync::mpsc::channel();
-        for _ in 0..10 {
-            let counter_clone = counter.clone();
-            let sender_clone = sender.clone();
-            std::thread::spawn(move || {
-                for _ in 0..10 {
-                    sender_clone.send(counter_clone.next()).unwrap();
-                }
-            });
-        }
-        receiver
-    };
-    let mut values: Vec<usize> = receiver.iter().collect();
-    values.sort_unstable();
-    assert_eq!((0_usize..100).collect::<Vec<usize>>(), values);
-}
-
 fn err_eq(a: &std::io::Error, b: &std::io::Error) -> bool {
     a.kind() == b.kind() && format!("{}", a) == format!("{}", b)
 }
@@ -245,6 +208,16 @@ impl From<StartThreadsError> for NewThreadPoolError {
         }
     }
 }
+impl From<NewThreadPoolError> for std::io::Error {
+    fn from(new_thread_pool_error: NewThreadPoolError) -> Self {
+        match new_thread_pool_error {
+            NewThreadPoolError::Parameter(s) => std::io::Error::new(ErrorKind::InvalidInput, s),
+            NewThreadPoolError::Spawn(s) => {
+                std::io::Error::new(ErrorKind::Other, format!("failed to start threads: {}", s))
+            }
+        }
+    }
+}
 
 #[derive(Debug)]
 pub enum TryScheduleError {
@@ -293,6 +266,29 @@ impl From<StartThreadsError> for TryScheduleError {
         }
     }
 }
+impl From<TryScheduleError> for std::io::Error {
+    fn from(try_schedule_error: TryScheduleError) -> Self {
+        match try_schedule_error {
+            TryScheduleError::QueueFull => {
+                std::io::Error::new(ErrorKind::WouldBlock, "TryScheduleError::QueueFull")
+            }
+            TryScheduleError::NoThreads(e) => std::io::Error::new(
+                e.kind(),
+                format!(
+                    "ThreadPool workers all panicked, failed starting replacement threads: {}",
+                    e
+                ),
+            ),
+            TryScheduleError::Respawn(e) => std::io::Error::new(
+                e.kind(),
+                format!(
+                    "ThreadPool failed starting threads to replace panicked threads: {}",
+                    e
+                ),
+            ),
+        }
+    }
+}
 
 struct Inner {
     name: &'static str,
@@ -302,11 +298,11 @@ struct Inner {
 }
 
 impl Inner {
-    pub fn num_live_threads(self: &Arc<Inner>) -> usize {
+    fn num_live_threads(self: &Arc<Self>) -> usize {
         Arc::strong_count(self) - 1
     }
 
-    fn work(self: &Arc<Inner>) {
+    fn work(self: &Arc<Self>) {
         loop {
             let recv_result = self
                 .receiver
@@ -334,9 +330,9 @@ impl Inner {
         name: String,
         f: impl FnOnce() + Send + 'static,
     ) -> Result<(), std::io::Error> {
-        // I found no way to make std::thread fail reliably on both macOS & Linux.
-        #[cfg(test)]
-        if num_live_threads >= TEST_MAX_THREADS.load(Ordering::Acquire) {
+        // I found no way to make std::thread fail reliably on both macOS & Linux.  So we simulate it with this.
+        #[cfg(feature = "testing")]
+        if num_live_threads >= INTERNAL_MAX_THREADS.load(std::sync::atomic::Ordering::Acquire) {
             return Err(std::io::Error::new(
                 std::io::ErrorKind::Other,
                 "err1".to_string(),
@@ -347,7 +343,7 @@ impl Inner {
         Ok(())
     }
 
-    fn start_thread(self: &Arc<Inner>) -> Result<(), StartThreadsError> {
+    fn start_thread(self: &Arc<Self>) -> Result<(), StartThreadsError> {
         let self_clone = self.clone();
         let num_live_threads = self.num_live_threads() - 1;
         if num_live_threads < self.size {
@@ -367,7 +363,7 @@ impl Inner {
         Ok(())
     }
 
-    fn start_threads(self: &Arc<Inner>) -> Result<(), StartThreadsError> {
+    fn start_threads(self: &Arc<Self>) -> Result<(), StartThreadsError> {
         while self.num_live_threads() < self.size {
             self.start_thread()?;
         }
@@ -500,6 +496,14 @@ impl ThreadPool {
         self.inner.num_live_threads()
     }
 
+    #[cfg(feature = "testing")]
+    #[doc(hidden)]
+    #[must_use]
+    pub fn num_live_threads_fn(&self) -> Box<dyn Fn() -> usize> {
+        let inner_clone = self.inner.clone();
+        Box::new(move || inner_clone.num_live_threads())
+    }
+
     /// Adds a job to the queue.  The next idle thread will execute it.
     /// Jobs are started in FIFO order.
     ///
@@ -590,138 +594,3 @@ impl Debug for ThreadPool {
         )
     }
 }
-
-#[cfg(test)]
-mod test {
-    use super::*;
-    use safe_lock::SafeLock;
-    use std::ops::Range;
-    use std::sync::atomic::{AtomicBool, Ordering};
-    use std::time::Instant;
-    static LOCK: SafeLock = SafeLock::new();
-
-    fn assert_elapsed(before: Instant, range_ms: Range<u64>) {
-        assert!(!range_ms.is_empty(), "invalid range {:?}", range_ms);
-        let elapsed = before.elapsed();
-        println!("elapsed = {} ms", elapsed.as_millis());
-        let duration_range =
-            Duration::from_millis(range_ms.start)..Duration::from_millis(range_ms.end);
-        assert!(
-            duration_range.contains(&elapsed),
-            "{:?} elapsed, out of range {:?}",
-            elapsed,
-            duration_range
-        );
-    }
-
-    fn set_max_threads(n: usize) {
-        println!("set_max_threads({})", n);
-        TEST_MAX_THREADS.store(n, Ordering::Release);
-    }
-
-    fn panic_threads(pool: &ThreadPool, num: usize) {
-        let pause = Arc::new(AtomicBool::new(true));
-        for _ in 0..num {
-            let pause_clone = pause.clone();
-            pool.try_schedule(move || {
-                println!(
-                    "thread {:?} waiting",
-                    std::thread::current().name().unwrap_or("")
-                );
-                while pause_clone.load(Ordering::Acquire) {
-                    sleep_ms(10);
-                }
-                println!(
-                    "panicking thread {:?}",
-                    std::thread::current().name().unwrap_or("")
-                );
-                panic!("ignore this panic");
-            })
-            .unwrap();
-        }
-        sleep_ms(100);
-        pause.store(false, Ordering::Release);
-        sleep_ms(100);
-    }
-
-    #[test]
-    fn new_thread_pool_at_max() {
-        let _guard = LOCK.lock().unwrap();
-        set_max_threads(2);
-        ThreadPool::new("test", 2).unwrap();
-    }
-
-    #[test]
-    fn new_thread_pool_error_spawn() {
-        let _guard = LOCK.lock().unwrap();
-        set_max_threads(2);
-        let err = ThreadPool::new("test", 3).unwrap_err();
-        assert!(matches!(err, NewThreadPoolError::Spawn(_)));
-    }
-
-    #[test]
-    fn schedule_retries_thread_start() {
-        let _guard = LOCK.lock().unwrap();
-        set_max_threads(3);
-        let pool = ThreadPool::new("test", 3).unwrap();
-        panic_threads(&pool, 3);
-        set_max_threads(0);
-        let before = Instant::now();
-        std::thread::spawn(|| {
-            sleep_ms(100);
-            set_max_threads(1);
-        });
-        let (sender, receiver) = std::sync::mpsc::channel();
-        pool.schedule(move || {
-            println!("sending ()");
-            sender.send(()).unwrap();
-        });
-        receiver.recv_timeout(Duration::from_millis(500)).unwrap();
-        assert_elapsed(before, 100..200);
-    }
-
-    // schedule_retries_when_queue_full is in tests/test.rs .
-
-    // try_schedule_queue_full is in tests/test.rs .
-
-    #[test]
-    fn try_schedule_no_threads() {
-        let _guard = LOCK.lock().unwrap();
-        set_max_threads(2);
-        let pool = ThreadPool::new("test", 2).unwrap();
-        panic_threads(&pool, 2);
-        set_max_threads(0);
-        let result = pool.try_schedule(|| {});
-        assert!(
-            matches!(result, Err(TryScheduleError::NoThreads(_))),
-            "{:?}",
-            result
-        );
-    }
-
-    #[test]
-    fn try_schedule_respawn() {
-        let _guard = LOCK.lock().unwrap();
-        set_max_threads(2);
-        let pool = ThreadPool::new("test", 2).unwrap();
-        panic_threads(&pool, 1);
-        set_max_threads(1);
-        let result = pool.try_schedule(|| {});
-        assert!(
-            matches!(result, Err(TryScheduleError::Respawn(_))),
-            "{:?}",
-            result
-        );
-    }
-
-    #[test]
-    fn threads_stop_after_pool_drops() {
-        let _guard = LOCK.lock().unwrap();
-        set_max_threads(2);
-        let pool = ThreadPool::new("test", 2).unwrap();
-        let inner = pool.inner.clone();
-        drop(pool);
-        std::thread::sleep(Duration::from_millis(100));
-        assert_eq!(0, inner.num_live_threads());
-    }
-}
